import { ResponseBody } from "./interface";
/**
 * Heimdal response
 *
 * @type {HeimdalResponse}
 */
export declare const HeimdalResponse: {
    new (serverUrl: string, responseObject: any, action?: string): {
        serverUrl: string;
        challenge: string;
        time: number;
        address: string;
        signature: string;
        fields: {
            [key: string]: string | number;
        };
        action: string;
        errors: any;
        signed: any;
        bap: any;
        isValid(): boolean;
        /**
         * Get the string message to sign for the verification
         *
         * This does not include the BAP data, as BAP also needs to include a signature of the data to be
         * validated (using this function), which creates a circular dependency. BAP is signed separately
         * and all the BAP data can be verified with on chain attestations.
         *
         * @returns {string}
         */
        getSigningMessage(): string;
        setSignature(address: string, signature: string): void;
        getResponseUrl(): string;
        getResponseBody(): ResponseBody | null;
        getId(): string;
        getChallenge(): string;
        getFields(): {
            [key: string]: string | number;
        };
        getFieldValue(fieldName: string): string | number | undefined;
        setFieldValue(fieldName: string, fieldValue: string | number): void;
        getVerifiedField(fieldName: string): any | undefined;
        getVerifiedFieldValue(fieldName: string): string | number | undefined;
        getVerifiedFieldNonce(fieldName: string): string | undefined;
    };
};
