import bsv, { Networks } from 'bsv';
import Message from 'bsv/message';
import moment from 'moment';
import base64url from 'base64url';

/******************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __classPrivateFieldGet(receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
}

function __classPrivateFieldSet(receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
}

// parseUri 1.2.2
// (c) Steven Levithan <stevenlevithan.com>
// MIT License
// http://blog.stevenlevithan.com/archives/parseuri
const parseUri = function (str) {
    var _a;
    /* eslint-disable max-len */
    const o = {
        strictMode: false,
        key: ['source', 'protocol', 'authority', 'userInfo', 'user', 'password', 'host', 'port', 'relative', 'path', 'directory', 'file', 'query', 'anchor'],
        q: {
            name: 'queryKey',
            parser: /(?:^|&)([^&=]*)=?([^&]*)/g,
        },
        parser: {
            strict: /^(?:([^:\/?#]+):)?(?:\/\/((?:(([^:@]*)(?::([^:@]*))?)?@)?([^:\/?#]*)(?::(\d*))?))?((((?:[^?#\/]*\/)*)([^?#]*))(?:\?([^#]*))?(?:#(.*))?)/,
            loose: /^(?:(?![^:@]+:[^:@\/]*@)([^:\/?#.]+):)?(?:\/\/)?((?:(([^:@]*)(?::([^:@]*))?)?@)?([^:\/?#]*)(?::(\d*))?)(((\/(?:[^?#](?![^?#\/]*\.[^?#\/.]+(?:[?#]|$)))*\/?)?([^?#\/]*))(?:\?([^#]*))?(?:#(.*))?)/,
        },
    };
    const m = o.parser['loose'].exec(str);
    const uri = {};
    let i = 14;
    while (i--) {
        if (m) {
            uri[o.key[i]] = m[i];
        }
    }
    uri[o.q.name] = {};
    (_a = uri[o.key[12]]) === null || _a === void 0 ? void 0 : _a.replace(o.q.parser, ($0, $1, $2) => {
        // @ts-ignore
        if ($1)
            uri[o.q.name][$1] = $2;
    });
    return uri;
};
// https://github.com/substack/json-stable-stringify/commit/e43ca2a1dcfc39bf1514684492767ef6040d1f3e
// MIT License
/* eslint-disable */
const jsonStableStringify = function (obj, opts) {
    if (!opts)
        opts = {};
    if (typeof opts === 'function')
        opts = { cmp: opts };
    let space = opts.space || '';
    if (typeof space === 'number') {
        space = Array(space + 1).join(' ');
    }
    const cycles = (typeof opts.cycles === 'boolean') ? opts.cycles : false;
    const replacer = opts.replacer || function (key, value) { return value; };
    const cmp = opts.cmp && (function (f) {
        return function (node) {
            return function (a, b) {
                const aobj = {
                    key: a,
                    value: node[a],
                };
                const bobj = {
                    key: b,
                    value: node[b],
                };
                return f(aobj, bobj);
            };
        };
    })(opts.cmp);
    const seen = [];
    return (function stringify(parent, key, node, level) {
        const indent = space ? ('\n' + new Array(level + 1).join(space)) : '';
        const colonSeparator = space ? ': ' : ':';
        if (node && node.toJSON && typeof node.toJSON === 'function') {
            node = node.toJSON();
        }
        node = replacer.call(parent, key, node);
        if (node === undefined) {
            return;
        }
        if (typeof node !== 'object' || node === null) {
            return JSON.stringify(node);
        }
        if (Array.isArray(node)) {
            const out = [];
            for (let i = 0; i < node.length; i++) {
                const item = stringify(node, i, node[i], level + 1) || JSON.stringify(null);
                out.push(indent + space + item);
            }
            return '[' + out.join(',') + indent + ']';
        }
        else {
            if (seen.indexOf(node) !== -1) {
                if (cycles)
                    return JSON.stringify('__cycle__');
                throw new TypeError('Converting circular structure to JSON');
            }
            else {
                seen.push(node);
            }
            const keys = Object.keys(node).sort(cmp && cmp(node));
            const out = [];
            for (let i = 0; i < keys.length; i++) {
                const k = keys[i];
                const value = stringify(node, k, node[k], level + 1);
                if (!value)
                    continue;
                const keyValue = JSON.stringify(k)
                    + colonSeparator
                    + value;
                out.push(indent + space + keyValue);
            }
            seen.splice(seen.indexOf(node), 1);
            return '{' + out.join(',') + indent + '}';
        }
    })({ '': obj }, '', obj, 0);
};

const DEFAULT_TYPE = 'api';
const DEFAULT_ACTION = '/loginViaQr';
const ALLOWED_TYPES = ['api', 'app', 'add', 'fetch', 'sign'];
const AIP_BITCOM_ADDRESS = '15PciHG22SNLQJXMoSUaWVi7WSqc7hCfva';

/**
 * Heimdal response
 *
 * @type {HeimdalResponse}
 */
const HeimdalResponse = class {
    constructor(serverUrl, responseObject, action = '') {
        // remove trailing slash if added by accident
        serverUrl = serverUrl.replace(/\/+$/, '');
        const { challenge, time, address, signature, fields, bap, } = responseObject;
        this.serverUrl = serverUrl;
        this.challenge = challenge;
        this.time = time;
        this.address = address;
        this.signature = signature;
        this.fields = fields || [];
        this.signed = null;
        this.bap = bap || [];
        // we need this from the heimdal object to create the response url
        this.action = action || DEFAULT_ACTION;
        this.errors = [];
    }
    isValid() {
        // pre-check
        if (!this.serverUrl || !this.challenge || !this.time || !this.address || !this.signature) {
            this.errors.push('Not all fields are filled in');
            return false;
        }
        const mTime = moment.unix(this.time);
        if (!mTime.isValid()) {
            this.errors.push('Invalid time');
            return false;
        }
        if (mTime.isAfter(moment())) {
            this.errors.push('Time is in the future');
            return false;
        }
        if (mTime.add(30, 'seconds').isBefore(moment())) {
            this.errors.push('Signature has expired');
            return false;
        }
        // check the signature from this user and validate the public key
        // server url includes the protocol, but not the trailing slash - https://demo.heimdal.app
        const messageBuffer = Buffer.from(this.getSigningMessage());
        try {
            if (!Message.verify(messageBuffer, this.address, this.signature)) {
                this.errors.push('Could not verify signature');
                return false;
            }
        }
        catch (e) {
            this.errors.push('Could not verify signature');
            return false;
        }
        if (this.bap && this.bap.length) {
            try {
                // BAP uses the same message for signing as the normal request
                if (!Message.verify(messageBuffer, this.bap.address, this.bap.signature)) {
                    this.errors.push('Could not verify BAP signature');
                    return false;
                }
            }
            catch (e) {
                this.errors.push('Could not verify BAP signature');
                return false;
            }
        }
        return true;
    }
    /**
     * Get the string message to sign for the verification
     *
     * This does not include the BAP data, as BAP also needs to include a signature of the data to be
     * validated (using this function), which creates a circular dependency. BAP is signed separately
     * and all the BAP data can be verified with on chain attestations.
     *
     * @returns {string}
     */
    getSigningMessage() {
        return this.serverUrl
            + '/'
            + this.challenge
            + '?time=' + this.time
            // @ts-ignore
            + '&f=' + encodeURIComponent(jsonStableStringify(this.fields));
    }
    setSignature(address, signature) {
        this.address = address;
        this.signature = signature;
    }
    getResponseUrl() {
        return `${this.serverUrl}${this.action}`;
    }
    getResponseBody() {
        if (!this.challenge || !this.time || !this.address || !this.signature) {
            this.errors.push('Missing one of the required fields');
            return null;
        }
        const responseBody = {
            challenge: this.challenge,
            time: this.time,
            address: this.address,
            signature: this.signature,
            fields: this.fields,
        };
        if (this.bap) {
            responseBody.bap = this.bap;
        }
        if (this.signed) {
            responseBody.signed = this.signed;
        }
        return responseBody;
    }
    getId() {
        return this.address;
    }
    getChallenge() {
        return this.challenge;
    }
    getFields() {
        return this.fields;
    }
    getFieldValue(fieldName) {
        return this.fields[fieldName];
    }
    setFieldValue(fieldName, fieldValue) {
        this.fields[fieldName] = fieldValue;
    }
    getVerifiedField(fieldName) {
        var _a;
        const verifiedFields = (_a = this.bap) === null || _a === void 0 ? void 0 : _a.attributes;
        if (verifiedFields) {
            return verifiedFields[fieldName];
        }
        return undefined;
    }
    getVerifiedFieldValue(fieldName) {
        const field = this.getVerifiedField(fieldName);
        if (field) {
            return field === null || field === void 0 ? void 0 : field.value;
        }
        return undefined;
    }
    getVerifiedFieldNonce(fieldName) {
        const field = this.getVerifiedField(fieldName);
        if (field) {
            return field === null || field === void 0 ? void 0 : field.nonce;
        }
        return undefined;
    }
};

const Random = {
    secret(length = 32) {
        return base64url(bsv.crypto.Random.getRandomBuffer(length)).substr(0, length);
    },
    hexString(length = 32) {
        return bsv.crypto.Random.getRandomBuffer(Math.ceil(length / 2)).toString('hex');
    },
};

var _privateKey, _address, _a;
/**
 * Heimdal
 *
 * @type {HeimdalId}
 */
const HeimdalId = (_a = class {
        constructor(privateKey) {
            _privateKey.set(this, void 0);
            _address.set(this, '');
            if (privateKey) {
                this.setPrivateKey(privateKey);
            }
            this.fields = [];
            this.resetVariables();
        }
        /**
         * Set, or re-set, the private key to use during operations
         *
         * @param privateKey
         */
        setPrivateKey(privateKey) {
            __classPrivateFieldSet(this, _privateKey, bsv.PrivateKey.fromWIF(privateKey), "f");
            __classPrivateFieldSet(this, _address, __classPrivateFieldGet(this, _privateKey, "f").toAddress(Networks.mainnet).toString(), "f");
        }
        getAddress() {
            return __classPrivateFieldGet(this, _address, "f");
        }
        /**
         * Reset all internal variables to their defaults
         */
        resetVariables() {
            this.source = '';
            this.protocol = 'heimdal';
            this.host = '';
            this.port = '';
            this.authority = '';
            this.challenge = '';
            this.parameters = {};
            this.type = DEFAULT_TYPE;
            this.action = DEFAULT_ACTION;
            this.fields = [];
            this.value = undefined;
            this.extension = '';
            this.signData = '';
            this.signature = undefined;
            this.id = undefined;
            this.errors = [];
        }
        /**
         * Start a new Heimdal request
         * @param authority
         */
        newRequest(authority) {
            if (!authority) {
                throw new Error('Authority is mandatory');
            }
            this.resetVariables();
            authority = authority.replace(/^https?:\/\//, ''); // remove protocol if applicable
            authority = authority.replace(/\/$/, ''); // remove trailing /
            const [host, port] = authority.split(':');
            this.host = host;
            this.port = port;
            this.authority = authority;
            this.challenge = Random.secret(32);
        }
        /**
         * Allow setting of custom challenge, if for instance it was created earlier
         *
         * @param challenge
         */
        setChallenge(challenge) {
            this.challenge = challenge;
        }
        getChallenge() {
            return this.challenge;
        }
        /**
         * Set the type of response requested, see ALLOWED_TYPES
         *
         * @param type
         */
        setType(type) {
            if (!ALLOWED_TYPES.includes(type)) {
                throw new Error('This type is not allowed');
            }
            this.type = type;
        }
        /**
         * Set the action to perform when returning the login response
         *
         * For web (api) - always add as absolute url (with leading "/")
         * For mobile (app) - give the name of the intent
         *
         * @param action
         */
        setAction(action) {
            this.action = action;
        }
        getChecksum(url) {
            const qrHex = bsv.crypto.Hash.sha256(Buffer.from(url));
            // @ts-ignore
            const address = bsv.PrivateKey.fromHex(qrHex)
                .publicKey
                .toAddress()
                .toString();
            return `${address.substr(-8, 4)}-${address.substr(-4)}`;
        }
        addField(fieldName) {
            this.fields.push(fieldName);
        }
        addFields(fields) {
            fields.forEach((fieldName) => {
                if (!this.fields.includes(fieldName)) {
                    this.fields.push(fieldName);
                }
            });
        }
        getFields() {
            return this.fields;
        }
        /**
         * This function cleans the fields of any /\*$/ present, indicating an optional field
         * @returns {[]}
         */
        getCleanFields() {
            return this.getFields().map((f) => { return f.replace('*', ''); });
        }
        setFields(fields) {
            this.fields = Array.isArray(fields) ? fields : [];
        }
        getValue() {
            return this.value;
        }
        getAction() {
            return this.action || '/loginViaQr';
        }
        getType() {
            return this.type || 'api';
        }
        getAuthority() {
            return this.authority || '';
        }
        getServerUrl() {
            return 'https://' + this.authority;
        }
        getId() {
            return this.id || null;
        }
        getSignature() {
            return this.signature || null;
        }
        getSigningData() {
            return this.signData || null;
        }
        setSigningData(signData) {
            this.signData = signData;
        }
        /**
         * Add a field value to the request
         *
         * This will overwrite any other fields set for this request + set the type to 'add' and unset action
         *
         * @param fieldName
         * @param fieldValue
         */
        addFieldValue(fieldName, fieldValue) {
            this.fields = [fieldName];
            this.value = fieldValue;
            this.setType('add');
            this.setAction('');
        }
        /**
         * Get the heimdal login url with the set parameters
         *
         * @param short {boolean} Whether to return a url that is as short as possible, with default values
         *
         * @throws Error on missing authority ot challenge
         * @returns {string}
         */
        getRequest(short = false) {
            if (!this.authority || !this.challenge) {
                throw new Error('Not initialized properly');
            }
            let url = this.protocol + '://'
                + this.authority
                + '/'
                + this.challenge;
            let first = true;
            if (!short || this.type !== DEFAULT_TYPE) {
                url += (first ? '?' : '&') + 't=' + this.type;
                first = false;
            }
            if (!short || this.action !== DEFAULT_ACTION) {
                url += (first ? '?' : '&') + 'a=' + this.action;
                first = false;
            }
            if (this.fields.length) {
                url += (first ? '?' : '&') + 'f=' + this.fields.map((f) => {
                    return encodeURIComponent(f);
                })
                    .join(',');
                first = false;
            }
            if (this.value) {
                url += (first ? '?' : '&') + 'v=' + encodeURIComponent(this.value);
            }
            return url;
        }
        /**
         * Get a signed request url
         *
         * @param short
         *
         * @throws Error on missing private key
         * @returns {string}
         */
        getSignedRequest(short = false) {
            if (!__classPrivateFieldGet(this, _privateKey, "f")) {
                throw new Error('No private key set for signing');
            }
            let request = this.getRequest(short);
            const message = this.getSigningMessage();
            const signature = this.signMessage(message);
            const first = !request.match(/\?/);
            request += (first ? '?' : '&') + 'sig=' + encodeURIComponent(signature);
            request += '&id=' + __classPrivateFieldGet(this, _address, "f");
            return request;
        }
        /**
         * Initialize a new request from a url
         *
         * @throws Error on illegal characters in url, not a valid Heimdal protocol string, missing authority
         * @param url
         */
        requestFromUrl(url) {
            var _a, _b, _c, _d, _e, _f, _g, _h;
            // check url for illegal characters - stop processing if found
            // https://tc39.es/ecma262/#sec-encodeuricomponent-uricomponent
            if (!url.match(/^[;,/?:@&=+$-_.!~*'%()a-z0-9]+$/)) {
                throw new Error('Illegal characters found in QR Code');
            }
            const parsedUrl = parseUri(url);
            if (!parsedUrl.protocol || parsedUrl.protocol !== 'heimdal') {
                throw new Error('Not a valid protocol for Heimdal');
            }
            if (!parsedUrl.authority) {
                throw new Error('Domain authority could not be parsed');
            }
            this.checksum = this.getChecksum(url);
            this.source = parsedUrl.source;
            this.protocol = parsedUrl.protocol;
            this.host = parsedUrl.host;
            this.port = parsedUrl.port;
            this.authority = parsedUrl.authority;
            const path = parsedUrl.path;
            this.challenge = path === null || path === void 0 ? void 0 : path.substr(1);
            this.parameters = parsedUrl.queryKey;
            // type and action should always reflect the url that was posted
            this.type = ((_a = this.parameters) === null || _a === void 0 ? void 0 : _a.t) || '';
            this.action = ((_b = this.parameters) === null || _b === void 0 ? void 0 : _b.a) || '';
            if ((_c = this.parameters) === null || _c === void 0 ? void 0 : _c.hasOwnProperty('v'))
                this.value = decodeURIComponent((_d = this.parameters) === null || _d === void 0 ? void 0 : _d.v);
            if ((_e = this.parameters) === null || _e === void 0 ? void 0 : _e.id)
                this.id = this.parameters.id;
            if ((_f = this.parameters) === null || _f === void 0 ? void 0 : _f.sig)
                this.signature = decodeURIComponent((_g = this.parameters) === null || _g === void 0 ? void 0 : _g.sig);
            if ((_h = this.parameters) === null || _h === void 0 ? void 0 : _h.f) {
                const f = this.parameters.f;
                this.fields = f.split(',')
                    .map((f) => {
                    return decodeURIComponent(f);
                });
                this.fields.sort();
            }
        }
        /**
         * Handle a new response from a request
         *
         * The response object should contain these fields:
         *   challenge - the challenge string from the request
         *   time - the time the user signed the message
         *   address - the address corresponding to the signing key
         *   signature - the signature in base64
         *   fields - any data fields that were sent across
         *
         * @param serverUrl
         * @param responseObject
         * @param action
         */
        newResponse(serverUrl, responseObject, action = '') {
            return new HeimdalResponse(serverUrl, responseObject, action || this.action || DEFAULT_ACTION);
        }
        /**
         * Create a new response object from the loaded parameters
         *
         * This can be used to send a response back to a server for verification for a login
         *
         * @param fields {array}
         */
        createResponse(fields) {
            const response = new HeimdalResponse(this.getServerUrl(), {
                challenge: this.getChallenge(),
                time: moment().unix(),
                fields,
            }, this.action || DEFAULT_ACTION);
            if (__classPrivateFieldGet(this, _privateKey, "f")) {
                // create a signature from the loaded private key
                const message = response.getSigningMessage();
                response.signature = this.signMessage(message);
                response.address = this.getAddress() || '';
                if (this.signData) {
                    response.signed = this.getSignedData();
                }
            }
            return response;
        }
        /**
         * Verify whether the heimdal url is valid and has been signed properly (if applicable)
         *
         * @return boolean
         */
        verifyRequest() {
            if (this.protocol !== 'heimdal')
                return false;
            if (!this.authority)
                return false;
            if (this.signature) {
                const message = this.getSigningMessage();
                return this.verifyMessage(message, this.id, this.signature);
            }
            return true;
        }
        /**
         * Get the message string to sign with the private key
         *
         * @throws Error on missing authority ot challenge
         * @return string
         */
        getSigningMessage() {
            if (!this.authority) {
                throw new Error('Domain authority has not been set');
            }
            if (!this.challenge) {
                throw new Error('Challenge key has not been set');
            }
            const fields = this.fields.map((f) => {
                return encodeURIComponent(f);
            });
            fields.sort();
            return this.protocol + '://'
                + this.authority
                + '/'
                + this.challenge
                + '?t=' + (this.type || DEFAULT_TYPE)
                + '&a=' + (this.action || DEFAULT_ACTION)
                + '&f=' + fields.join(',')
                + '&v=' + (this.value || '')
                + '&x=' + (this.extension || '');
        }
        /**
         * Sign the set signData with the correct function
         */
        getSignedData() {
            var _a, _b, _c;
            if ((_a = this.signData) === null || _a === void 0 ? void 0 : _a.message) {
                return {
                    message: this.signData.message,
                    signature: this.signMessage(this.signData.message),
                    address: __classPrivateFieldGet(this, _address, "f"),
                };
            }
            if ((_b = this.signData) === null || _b === void 0 ? void 0 : _b.op_return) {
                return {
                    algorithm: 'AIP',
                    op_return: this.signOpReturn(this.signData.op_return),
                };
            }
            if ((_c = this.signData) === null || _c === void 0 ? void 0 : _c.tx) {
                return this.signTx(this.signData.tx);
            }
            return false;
        }
        /**
         * Sign the given op return data array
         *
         * @param opReturn
         */
        signOpReturn(opReturn) {
            const aipMessageBuffer = this.getAIPMessageBuffer(opReturn);
            // TODO is this correct ???
            const { 
            // @ts-ignore
            address, 
            // @ts-ignore
            signature, } = this.signMessage(aipMessageBuffer);
            return opReturn.concat([
                Buffer.from('|').toString('hex'),
                Buffer.from(AIP_BITCOM_ADDRESS).toString('hex'),
                Buffer.from('BITCOIN_ECDSA').toString('hex'),
                Buffer.from(address).toString('hex'),
                Buffer.from(signature, 'base64').toString('hex'),
            ]);
        }
        /**
         * get Message buffer for AIP
         *
         * @param opReturn
         * @returns {Buffer}
         */
        getAIPMessageBuffer(opReturn) {
            const buffers = [];
            if (opReturn[0].replace('0x', '') !== '6a') {
                // include OP_RETURN in constructing the signature buffer
                buffers.push(Buffer.from('6a', 'hex'));
            }
            opReturn.forEach((op) => {
                buffers.push(Buffer.from(op.replace('0x', ''), 'hex'));
            });
            // add a trailing "|" - this is the AIP way
            buffers.push(Buffer.from('|'));
            return Buffer.concat([...buffers]);
        }
        /**
         * Sign the given Bitcoin transaction
         *
         * @param tx
         *
         * @return boolean
         */
        signTx(tx) {
            // TODO
            console.log(tx);
            // const transaction = bsv.Transaction.fromString(tx);
            return false;
        }
        /**
         * Sign a message with the private key of this instance
         *
         * @param message
         *
         * @return Signature Messages signature object ( { address, signature } )
         */
        signMessage(message) {
            const messageBuffer = Buffer.isBuffer(message) ? message : Buffer.from(message);
            return Message(messageBuffer).sign(__classPrivateFieldGet(this, _privateKey, "f"));
        }
        /**
         * Verify the message given against the signature given
         *
         * @param message {string} String message to verify the signature for
         * @param address {string} Bitcoin address
         * @param signature {string} Signature in base64
         *
         * @return boolean
         */
        verifyMessage(message, address, signature) {
            try {
                return Message.verify(Buffer.from(message), address, signature);
            }
            catch (e) {
                return false;
            }
        }
    },
    _privateKey = new WeakMap(),
    _address = new WeakMap(),
    _a);

export { HeimdalId };
//# sourceMappingURL=typescript-npm-package.esm.js.map
